import os

from pyspark import Row
from pyspark.sql.connect.dataframe import DataFrame

from davidkhala.spark.sink.stream import ForeachBatchWriter


class NewRelic(ForeachBatchWriter):
    def __init__(self, license_key=None):
        self.license_key = license_key or os.environ.get('NEW_RELIC_LICENSE_KEY')

    @property
    def on_batch(self):
        s = self.license_key  # cloned string for spark serialize

        def func(df: DataFrame, batch_id: int):
            from davidkhala.newrelic.log import Ingestion
            i = Ingestion(s)
            data = {
                'data': df.toPandas().to_dict(),
                'batch_id': batch_id
            }

            i.send(str(data))

        return func

    @property
    def on_row(self):
        s = self.license_key  # cloned string for spark serialize

        def func(row: Row):
            from davidkhala.newrelic.log import Ingestion
            i = Ingestion(s)
            data = row.asDict(),
            i.send(str(data))

        return func



from abc import abstractmethod, ABC
from typing import Callable, Protocol

from pyspark.sql.connect.dataframe import DataFrame
from pyspark.sql.connect.session import SparkSession

from pyspark.sql.types import Row


class ForeachWriter(ABC):
    """
    object class used in DataStreamWriter.foreach(...)
    """

    def open(self, partition_id: int, epoch_id: int) -> bool: pass

    @abstractmethod
    def process(self, row: Row) -> None: ...

    def close(self, error: Exception | None) -> None: pass


class ForeachBatchWriter(Protocol):
    @property
    def on_batch(self) -> Callable[[DataFrame, int], None]: ...



from pyspark.sql.connect.dataframe import DataFrame
from pyspark.sql.connect.readwriter import DataFrameWriter
from pyspark.sql.connect.session import SparkSession


class Write:
    def __init__(self, df: DataFrame):
        assert not df.isStreaming
        self.batch: DataFrameWriter = df.write

    @property
    def spark(self) -> SparkSession:
        return self.batch._spark



from davidkhala.spark.sink.stream import ForeachBatchWriter
from pyspark.sql.connect.session import SparkSession
from pyspark.sql.connect.dataframe import DataFrame
from pyspark.sql.connect.streaming import StreamingQuery, DataStreamWriter


def startAny(df: DataFrame, writer: ForeachBatchWriter) -> StreamingQuery:
    assert df.isStreaming

    return df.writeStream.foreachBatch(writer.on_batch).start()


class Write:
    def __init__(self, df: DataFrame):
        assert df.isStreaming
        self.stream: DataStreamWriter = df.writeStream

    @property
    def spark(self) -> SparkSession:
        return self.stream._session



from datetime import datetime

from pyspark import SparkContext, SparkConf


class Wrapper(SparkContext):
    sc: SparkContext

    def __init__(self, sc: SparkContext):
        self.sc = sc

    def __getattr__(self, name):
        # Delegate unknown attributes/methods to the wrapped instance
        return getattr(self.sc, name)

    def disconnect(self):
        self.sc.stop()

    @property
    def startTime(self):
        return datetime.fromtimestamp(self.sc.startTime / 1000)

    @property
    def appTime(self) -> int:
        """
        assume local spark app, not YARN
        :return: nanoseconds since unix epoch
        """
        assert self.sc.applicationId.startswith('local-')
        epoch_nano = int(self.sc.applicationId[6:])
        assert epoch_nano > self.sc.startTime
        return epoch_nano

    @staticmethod
    def from_config(conf: SparkConf = SparkConf()):
        return Wrapper(SparkContext.getOrCreate(conf))



from pyspark.sql.connect.dataframe import DataFrame
from pyspark.sql.connect.session import SparkSession


def sample(spark: SparkSession, density=1) -> DataFrame:
    return spark.readStream.format('rate').option("rowsPerSecond", density).load()



from enum import Enum
from typing import Union

from pyspark.sql.connect.dataframe import DataFrame

from davidkhala.spark.sink import Write


class Mode(Enum):
    append = "append"
    overwrite = "overwrite"
    ignore = "ignore"
    error = "errorifexists"  # default. Throw an exception if data already exists.


class SQLServer(Write):

    def __init__(self, df: DataFrame, *,
                 server: str, database: str, table: str, user: str, password: str,
                 mode: Union[Mode,str] = Mode.error
                 ):
        super().__init__(df)
        self.url = f"jdbc:sqlserver://{server};databaseName={database}"
        self.table = table
        self.mode = mode.value if isinstance(mode,Mode) else Mode(mode).value
        self.properties = {
            "user": user,
            "password": password,
            "driver": "com.microsoft.sqlserver.jdbc.SQLServerDriver"
        }

    def start(self):
        self.batch.jdbc(
            self.url, self.table, self.mode, self.properties
        )



from pyspark import SparkConf
from pyspark.errors import IllegalArgumentException
from pyspark.sql import SparkSession as JavaSparkSession
from pyspark.sql.connect.session import SparkSession


class Wrapper:
    spark: SparkSession | JavaSparkSession

    def __init__(self, spark):
        self.spark = spark

    def disconnect(self):
        if not self.spark.is_stopped:
            self.spark.stop()

    @property
    def schema(self) -> str:
        """
        :return: current schema full name
        """
        return self.spark.catalog.currentCatalog() + '.' + self.spark.catalog.currentDatabase()

    def __getattr__(self, name):
        # Delegate unknown attributes/methods to the wrapped instance
        return getattr(self.spark, name)


class ServerMore(Wrapper):
    @property
    def appName(self):
        try:
            return self.spark.conf.get("spark.app.name")
        except IllegalArgumentException as e:
            if str(e).splitlines()[0] == "The value of property spark.app.name must not be null":
                return
            else:
                raise e

    @property
    def clusterId(self):
        """almost abstract method"""
        ...


def regular(*, name: str = None, conf: SparkConf = SparkConf()) -> JavaSparkSession:
    """
    Visit https://spark.apache.org/docs/latest/sql-getting-started.html#starting-point-sparksession for creating regular Spark Session
    """
    _ = JavaSparkSession.builder.config(conf=conf)
    if name: _.appName(name)

    return _.getOrCreate()


class Databricks(ServerMore):
    spark: SparkSession
    cluster_id: str

    def __init__(self, workspace_instance_name: str, token: str, cluster_id: str):
        from getpass import getuser
        user_id = getuser()  # can be any
        connection_string = f"sc://{workspace_instance_name}:443/;token={token};x-databricks-cluster-id={cluster_id};user_id={user_id}"
        session = SparkSession.builder.remote(connection_string).getOrCreate()
        super().__init__(session)
        self.cluster_id = cluster_id
        self.user_id = user_id

    @property
    def clusterId(self):
        return self.cluster_id
